open import Data.Product

x-comm : (A B : Set) → (A × B) → (B × A)
x-comm A B (fst , snd) = snd , fst

postulate A B C : Set

{- 3 : Implication -}
id : A → A
id a = a

K : A → B → A
K p = λ x → p

app : (A → B) → A → B
app p = p

flip : (A → B → C) → B → A → C
flip p = λ x x₁ → p x₁ x

comp : (A → B) → (B → C) → (A → C)
comp p = λ x x₁ → x (p x₁)

S : (A → B → C) → (A → B) → A → C
S x y z = x z (y z)

{- 4 : Conjection -}
proj1 : (A B : Set) →  (A × B) → A
proj1 A B (fst , snd) = fst

proj2 :(A B : Set) → (A × B) → B
proj2 A B (fst , snd) = snd

diag : A → (A × A)
diag = λ x → x , x

curry1 : (A × B → C) → (A → B → C)
curry1 f a b = f (a , b)

curry2 : (A → B → C) → (A × B → C)
curry2 f (fst , snd) = f fst snd

_↔_ : (A B : Set) → Set
_↔_ A B = (A → B) × (B → A)

currying : (A × B → C) ↔ (A → B → C)
currying = curry1 , curry2

distrib : (A → (B × C)) ↔ ((A → B) × (A → C))
distrib = (λ x → (λ x₁ → proj1 B C  (x x₁)) , λ x₁ → proj2 B C (x x₁)) , λ x x₁ → ((proj1 _ _ x) x₁ ) , ((proj2 _ _ x)x₁)

{- 5 : Disjunction -}

data _∨_ (A B : Set) : Set where
  left  : A → A ∨ B
  right : B → A ∨ B

or-elim : {A B C : Set} → (A ∨ B) → (A → C) → (B → C) → C
or-elim (left x) = λ x₁ x₂ → x₁ x
or-elim (right x) = λ x₁ x₂ → x₂ x


or-comm : {A B : Set}  → (A ∨ B) → (B ∨ A)
or-comm (left x) = right x
or-comm (right x) = left x

dist : {A B C : Set} → (A × (B ∨ C)) → (A × B) ∨ (A × C)
dist (fst , left x) = left (fst , x)
dist (fst , right x) = right (fst , x)

{- 6 : Negation -}

data ⊥ : Set where

⊥-elim : {A : Set} → ⊥ → A
⊥-elim ()

¬ : Set → Set
¬ = λ x → x → ⊥

contr : {A B : Set} → (A → B) → (¬ B → ¬ A)
contr f = λ x x₁ → x (f x₁)

non-contr : {A : Set} → ¬ (A × ¬ A)
non-contr = λ x → (proj2 _ _ x)(proj1 _ _ x)

nni : {A : Set} → A → ¬ (¬ A)
nni = λ x x₁ → x₁ x

⊥-nne : ¬ (¬ ⊥) → ⊥
⊥-nne = λ x → x λ x₁ → x₁

¬-elim : {A B : Set} → ¬ A → A → B
¬-elim = λ x x₁ → ⊥-elim (x x₁)


dm∧ : {A B : Set} → ¬ A × ¬ B → ¬ (A ∨ B)
dm∧ x (left x₁) = (proj1 _ _ x) x₁
dm∧ x (right x₁) = (proj2 _ _ x) x₁

conv-dm : {A B : Set} → ¬ (A ∨ B) → ¬ A × ¬ B
conv-dm = λ x → (λ x₁ → x (left x₁)) , (λ z → x (right z))


nnlem : {A : Set} → ¬ (¬ (A ∨ (¬ A)))
nnlem = λ x → x (right λ x₁ → x (left x₁))

rp : {A : Set} → (A → ¬ A) → ((¬ A) → A) → ⊥
rp = λ x y → x (y λ z → x z z) (y λ z → x z z)

{- 7 : Truth -}

data ⊤ : Set where
 true : ⊤

ti : {A : Set} → (⊤ → A) → A
ti = λ x → x true

dmnt : ¬ ⊤ → ⊥
dmnt = λ x → x true

dmtn : ⊥ → ¬ ⊤
dmtn = λ x y → x

{- 8 : Classical logic -}

lem : Set₁
lem = (A : Set) → A ∨ (¬ A)

nne : Set₁
nne = (A : Set) → ¬ (¬ A) → A

nne-lem : nne → lem
nne-lem = λ x A → x (A ∨ ((x : A) → ⊥)) (λ y → y (right λ z → y (left z)))

lem-nne' : (A : Set) → (A ∨ ¬ A) → ¬ (¬ A) → A
lem-nne' A (left x) y = x
lem-nne' A (right x) y = ⊥-elim (y x)

lem-nne : lem → nne
lem-nne x = λ A y → lem-nne' A (x A) y

pierce : Set₁
pierce = (A B : Set) → ((A → B) → A) → A

nne-pierce : nne → pierce
nne-pierce = λ x A B y → x A λ z → z (y λ a → ⊥-elim (z a))

pierce-lem : pierce → lem
pierce-lem = λ x A → x (A ∨ ((x : A) → ⊥)) ⊥ (λ y → right (λ a → y (left a)))

